from setuptools import setup

setup(name='Dot Plot',
      version='0.1.2.4',
      description="Nico's dot plotter",
      url='',
      author='N to the Icky',
      author_email='flyingcircus@example.com',
      license='',
      packages=['library'],
      zip_safe=False)
